import 'reflect-metadata';

var anyConverter = {
    deserialize: function (json) {
        return deepCopy(json);
    },
    serialize: function (obj) {
        return deepCopy(obj);
    },
};
function any() {
    return { converter: anyConverter };
}
function deepCopy(a) {
    return a;
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

var JsonConverterError = (function (_super) {
    __extends(JsonConverterError, _super);
    function JsonConverterError(message, segment, parent) {
        var _this = _super.call(this, message) || this;
        _this.segment = segment;
        _this.parent = parent;
        Object.setPrototypeOf(_this, JsonConverterError.prototype);
        return _this;
    }
    return JsonConverterError;
}(Error));

function tryParseBoolean(value) {
    switch (value.toLowerCase()) {
        case 'true':
        case 'yes':
        case '1':
            return true;
        case 'false':
        case 'no':
        case '0':
            return false;
        default:
            var errorMessage = 'Obj cannot be parsed to type <Boolean>';
            throw new JsonConverterError(errorMessage);
    }
}
function tryParseNumber(value) {
    var n = Number(value);
    if (isNaN(n)) {
        var errorMessage = 'Obj cannot be parsed to type <Number>';
        throw new JsonConverterError(errorMessage);
    }
    return n;
}
function isNullOrUndefined(obj) {
    return typeof obj === 'undefined' || obj === null;
}
function isString(obj) {
    return typeof obj === 'string' || obj instanceof String;
}
function isBoolean(obj) {
    return typeof obj === 'boolean' || obj instanceof Boolean;
}
function isNumber(obj) {
    return typeof obj === 'number' || obj instanceof Number;
}
function isObject(obj) {
    return Object(obj) === obj;
}

var booleanConverter = {
    deserialize: function (json, args, deserializeOptions) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        else if (isBoolean(json)) {
            return json;
        }
        else if (deserializeOptions.tryParse && isString(json)) {
            return tryParseBoolean(json);
        }
        else {
            var errorMessage = 'Fail to deserialize, expected type is <Boolean>, but json is not';
            throw new JsonConverterError(errorMessage);
        }
    },
    serialize: function (obj) {
        return obj;
    },
};

var numberConverter = {
    deserialize: function (json, args, deserializeOptions) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        else if (isNumber(json)) {
            return json;
        }
        else if (deserializeOptions.tryParse && isString(json)) {
            return tryParseNumber(json);
        }
        else {
            var errorMessage = 'Fail to deserialize, expected type is <Number>, but json is not';
            throw new JsonConverterError(errorMessage);
        }
    },
    serialize: function (obj) {
        return obj;
    },
};

var JSONCONVERTER_MAPPING = Symbol.for('tipify.mapping');
var globalSymbols = Object.getOwnPropertySymbols(global);
var mappingInitialized = (globalSymbols.indexOf(JSONCONVERTER_MAPPING) > -1);
if (!mappingInitialized) {
    global[JSONCONVERTER_MAPPING] = [];
}
var JsonConverterMapper = (function () {
    function JsonConverterMapper() {
    }
    JsonConverterMapper.getMapping = function () {
        return global[JSONCONVERTER_MAPPING];
    };
    JsonConverterMapper.createMappingForType = function (type) {
        var typeMapping = this.getMappingForType(type);
        if (!typeMapping) {
            typeMapping = {
                properties: [],
                type: type,
            };
            this.getMapping().push(typeMapping);
        }
        return typeMapping;
    };
    JsonConverterMapper.getMappingForType = function (type) {
        return this.getMapping().find(function (m) { return m.type === type; });
    };
    JsonConverterMapper.listMappingForExtendingType = function (type) {
        return this.getMapping().filter(function (m) { return m.type.prototype instanceof type; });
    };
    JsonConverterMapper.getDiscriminatorPropertyForTypeMapping = function (typeMapping) {
        if (typeMapping.options && typeMapping.options.discriminatorValue) {
            return;
        }
        var current = typeMapping;
        do {
            if (current.options && current.options.discriminatorProperty) {
                return current.options.discriminatorProperty;
            }
            current = current.parent;
        } while (current);
    };
    JsonConverterMapper.getAllPropertiesForTypeMapping = function (typeMapping) {
        var properties = [];
        var current = typeMapping;
        do {
            current.properties.forEach(function (property) {
                if (!properties.some(function (p) { return p.name === property.name; })) {
                    properties.push(property);
                }
            });
            current = current.parent;
        } while (current);
        return properties;
    };
    return JsonConverterMapper;
}());

var objectConverter = {
    deserialize: function (json, args, deserializeOptions) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        if (!isObject(json)) {
            var errorMessage = "Cannot deserialize to <" + args.type.name + ">, given json is not an object";
            throw new JsonConverterError(errorMessage);
        }
        var typeMapping = JsonConverterMapper.getMappingForType(args.type);
        if (!typeMapping) {
            var errorMessage = "Cannot get mapping <" + args.type.name + ">, "
                + 'this may occur when decorator @jsonObject is missing';
            throw new JsonConverterError(errorMessage);
        }
        var discriminatorProperty = JsonConverterMapper.getDiscriminatorPropertyForTypeMapping(typeMapping);
        if (discriminatorProperty) {
            var discriminatorValue_1 = json[discriminatorProperty];
            var subTypes = JsonConverterMapper.listMappingForExtendingType(args.type);
            var subType = subTypes.find(function (m) { return m.options && m.options.discriminatorValue === discriminatorValue_1; });
            if (!subType) {
                var errorMessage = "Polymorphism error : Cannot get subtype for <" + args.type.name + "> "
                    + ("got only subtypes <" + subTypes.map(function (t) { return t.options ? t.options.discriminatorValue : t.type.name; }).toString() + ">");
                throw new JsonConverterError(errorMessage);
            }
            return this.deserialize(json, { type: subType.type }, deserializeOptions);
        }
        var instance = new typeMapping.type();
        var properties = JsonConverterMapper.getAllPropertiesForTypeMapping(typeMapping);
        properties.forEach(function (property) {
            try {
                if (!deserializeOptions.keepObjectFieldValues || json.hasOwnProperty(property.serializedName)) {
                    instance[property.name] = property.converter.deserialize(json[property.serializedName], property.args, deserializeOptions);
                }
            }
            catch (err) {
                var errorMessage = "Fail to deserialize property <" + property.serializedName + ">";
                throw new JsonConverterError(errorMessage, property.serializedName, err);
            }
        });
        return instance;
    },
    serialize: function (obj, args, serializeOptions) {
        if (isNullOrUndefined(obj)) {
            return obj;
        }
        var typeMapping = JsonConverterMapper.getMappingForType(obj.constructor);
        var instance = {};
        if (!typeMapping && !serializeOptions.unsafe) {
            var errorMessage = "Cannot get mapping <" + obj.constructor.name + ">, "
                + 'this may occur when decorator @jsonObject is missing';
            throw new JsonConverterError(errorMessage);
        }
        if (typeMapping) {
            var properties = JsonConverterMapper.getAllPropertiesForTypeMapping(typeMapping);
            properties.forEach(function (property) {
                try {
                    var value = property.converter.serialize(obj[property.name], property.args, serializeOptions);
                    if (!isNullOrUndefined(value)) {
                        instance[property.serializedName] = value;
                    }
                }
                catch (err) {
                    var errorMessage = "Fail to serialize property <" + property.serializedName + ">";
                    throw new JsonConverterError(errorMessage, property.serializedName, err);
                }
            });
        }
        else {
            Object.keys(obj).forEach(function (property) {
                try {
                    if (!isNullOrUndefined(obj[property])) {
                        var converterAndArgs = autodetectConverterAndArgs(obj[property]);
                        var value = converterAndArgs.converter.serialize(obj[property], converterAndArgs.args, serializeOptions);
                        if (!isNullOrUndefined(value)) {
                            instance[property] = value;
                        }
                    }
                }
                catch (err) {
                    var errorMessage = "Fail to serialize property <" + property + ">";
                    throw new JsonConverterError(errorMessage, property, err);
                }
            });
        }
        return instance;
    },
};

var stringConverter = {
    deserialize: function (json) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        if (!isString(json)) {
            var errorMessage = 'Fail to deserialize string, expected type is <String>, but json is not';
            throw new JsonConverterError(errorMessage);
        }
        return json;
    },
    serialize: function (obj) {
        return obj;
    },
};

function normalizeConverterAndArgs(typeOrConverter) {
    var converter = typeOrConverter;
    if (converter
        && typeof converter.serialize === 'function'
        && typeof converter.deserialize === 'function') {
        return { converter: converter };
    }
    var converterWithArgs = typeOrConverter;
    if (converterWithArgs
        && converterWithArgs.converter
        && typeof converterWithArgs.converter.serialize === 'function'
        && typeof converterWithArgs.converter.deserialize === 'function') {
        return converterWithArgs;
    }
    if (typeOrConverter && Array.isArray(typeOrConverter) && typeOrConverter.length) {
        return arrayOf(typeOrConverter[0]);
    }
    var type = typeOrConverter;
    switch (type) {
        case String:
            return { converter: stringConverter };
        case Boolean:
            return { converter: booleanConverter };
        case Number:
            return { converter: numberConverter };
        default:
            return { converter: objectConverter, args: { type: type } };
    }
}
function autodetectConverterAndArgs(obj) {
    if (Array.isArray(obj)) {
        return { converter: arrayConverter, args: {} };
    }
    else if (isString(obj)) {
        return { converter: stringConverter };
    }
    else if (isBoolean(obj)) {
        return { converter: booleanConverter };
    }
    else if (isNumber(obj)) {
        return { converter: numberConverter };
    }
    else if (isObject(obj)) {
        return { converter: objectConverter, args: { type: obj.constructor } };
    }
    else {
        var errorMessage = 'Cannot get detect type of given obj';
        throw new JsonConverterError(errorMessage);
    }
}

var arrayConverter = {
    deserialize: function (json, args, deserializeOptions) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        if (!Array.isArray(json)) {
            var errorMessage = 'Fail to deserialize, given json is not an array';
            throw new JsonConverterError(errorMessage);
        }
        var arr = [];
        json.forEach(function (entry, index) {
            try {
                arr[index] = args.itemConverter.deserialize(entry, args.itemConverterArgs, deserializeOptions);
            }
            catch (err) {
                var errorMessage = "Fail to deserialize index <" + index + "> of array";
                throw new JsonConverterError(errorMessage, index, err);
            }
        });
        return arr;
    },
    serialize: function (obj, args, serializeOptions) {
        if (isNullOrUndefined(obj)) {
            return obj;
        }
        if (!Array.isArray(obj)) {
            var errorMessage = 'Fail to serialize, given json is not an array';
            throw new JsonConverterError(errorMessage);
        }
        var arr = [];
        obj.forEach(function (entry, index) {
            try {
                if (!entry) {
                    arr[index] = entry;
                }
                else {
                    var converterAndArgs = args.itemConverter
                        ? { converter: args.itemConverter, args: args.itemConverterArgs }
                        : autodetectConverterAndArgs(entry);
                    arr[index] = converterAndArgs.converter.serialize(entry, converterAndArgs.args, serializeOptions);
                }
            }
            catch (err) {
                var errorMessage = "Fail to serialize index <" + index + "> of array";
                throw new JsonConverterError(errorMessage, index, err);
            }
        });
        return arr;
    },
};
function arrayOf(itemTypeOrConverter) {
    var _a = normalizeConverterAndArgs(itemTypeOrConverter), converter = _a.converter, args = _a.args;
    return {
        args: { itemConverter: converter, itemConverterArgs: args },
        converter: arrayConverter,
    };
}

var enumConverter = {
    deserialize: function (json, args) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        if (isNullOrUndefined(args.type[json])) {
            var errorMessage = "Enum value <" + json + "> cannot be found in [" + Object.keys(args.type) + "]";
            throw new JsonConverterError(errorMessage);
        }
        if (isNumber(json)) {
            return json;
        }
        return args.type[json];
    },
    serialize: function (obj, args) {
        if (isNullOrUndefined(obj)) {
            return obj;
        }
        if (isNullOrUndefined(args.type[obj])) {
            var errorMessage = "Enum value <" + obj + "> cannot be found";
            throw new JsonConverterError(errorMessage);
        }
        if (args.strategy === EnumStrategy.INDEX) {
            return obj;
        }
        else if (args.strategy === EnumStrategy.NAME) {
            return args.type[obj];
        }
        else {
            var errorMessage = 'Enum strategy is not defined';
            throw new JsonConverterError(errorMessage);
        }
    },
};
var EnumStrategy;
(function (EnumStrategy) {
    EnumStrategy[EnumStrategy["INDEX"] = 0] = "INDEX";
    EnumStrategy[EnumStrategy["NAME"] = 1] = "NAME";
})(EnumStrategy || (EnumStrategy = {}));
function enumOf(type, strategy) {
    if (strategy === void 0) { strategy = EnumStrategy.NAME; }
    return { converter: enumConverter, args: { type: type, strategy: strategy } };
}

var keyValueConverter = {
    deserialize: function (json, args, deserializeOptions) {
        if (isNullOrUndefined(json)) {
            return json;
        }
        if (!isObject(json) || Array.isArray(json)) {
            var errorMessage = 'Fail to deserialize map, given json is not an object';
            throw new JsonConverterError(errorMessage);
        }
        var instance = {};
        Object.keys(json).forEach(function (k) {
            var key;
            if (args.keyType === String) {
                key = String(k);
            }
            else {
                key = Number(k);
                if (Number.isNaN(key)) {
                    var errorMessage = "Fail to parse map key <" + k + "> should be a <" + args.keyType.name + ">";
                    throw new JsonConverterError(errorMessage);
                }
            }
            try {
                instance[key] = args.valueConverter.deserialize(json[key], args.valueConverterArgs, deserializeOptions);
            }
            catch (err) {
                var errorMessage = "Fail to deserialize map value with key <" + key + ">";
                throw new JsonConverterError(errorMessage, key, err);
            }
        });
        return instance;
    },
    serialize: function (obj, args, serializeOptions) {
        if (isNullOrUndefined(obj)) {
            return obj;
        }
        var instance = {};
        Object.keys(obj).forEach(function (key) {
            try {
                instance[key] = args.valueConverter.serialize(obj[key], args.valueConverterArgs, serializeOptions);
            }
            catch (err) {
                var errorMessage = "Fail to serialize map value with key <" + key + ">";
                throw new JsonConverterError(errorMessage, key, err);
            }
        });
        return instance;
    },
};
function keyValueOf(keyType, valueTypeOrConverter) {
    var _a = normalizeConverterAndArgs(valueTypeOrConverter), converter = _a.converter, args = _a.args;
    return {
        args: { keyType: keyType, valueConverter: converter, valueConverterArgs: args },
        converter: keyValueConverter,
    };
}

var JsonConverter = (function () {
    function JsonConverter(_options) {
        if (_options === void 0) { _options = DEFAULT_JSON_CONVERTER_OPTIONS; }
        this._options = _options;
    }
    JsonConverter.prototype.serialize = function (obj, type, options) {
        if (options === void 0) { options = this._options.serialize || DEFAULT_SERIALIZE_OPTIONS; }
        if (isNullOrUndefined(obj)) {
            return obj;
        }
        try {
            var _a = type ? normalizeConverterAndArgs(type) : autodetectConverterAndArgs(obj), converter = _a.converter, args = _a.args;
            return converter.serialize(obj, args, options);
        }
        catch (err) {
            var errorMessage = this.buildErrorMessage('Fail to serialize json at ', err);
            throw new JsonConverterError(errorMessage);
        }
    };
    JsonConverter.prototype.deserialize = function (json, type, options) {
        if (options === void 0) { options = this._options.deserialize || DEFAULT_DESERIALIZE_OPTIONS; }
        if (isNullOrUndefined(json)) {
            return json;
        }
        try {
            if (!type) {
                var errorMessage = 'Type is required';
                throw new JsonConverterError(errorMessage);
            }
            var _a = normalizeConverterAndArgs(type), converter = _a.converter, args = _a.args;
            return converter.deserialize(json, args, options);
        }
        catch (err) {
            var errorMessage = this.buildErrorMessage('Fail to deserialize json at ', err);
            throw new JsonConverterError(errorMessage);
        }
    };
    JsonConverter.prototype.buildErrorMessage = function (message, err) {
        var errorMessage = '';
        var segment = '/';
        while (err) {
            errorMessage += '\n    -> ' + err.message;
            if (err instanceof JsonConverterError) {
                if (!isNullOrUndefined(err.segment)) {
                    segment += '/' + err.segment;
                }
                err = err.parent;
            }
            else {
                break;
            }
        }
        return message + segment + errorMessage;
    };
    return JsonConverter;
}());
var DEFAULT_SERIALIZE_OPTIONS = {
    unsafe: false,
};
var DEFAULT_DESERIALIZE_OPTIONS = {
    keepObjectFieldValues: true,
    tryParse: false,
};
var DEFAULT_JSON_CONVERTER_OPTIONS = {
    deserialize: DEFAULT_DESERIALIZE_OPTIONS,
    serialize: DEFAULT_SERIALIZE_OPTIONS,
};

function jsonObject(options) {
    return function (constructor) {
        var mapping = JsonConverterMapper.createMappingForType(constructor);
        mapping.options = options;
        var parentType = Object.getPrototypeOf(constructor);
        if (parentType.name) {
            mapping.parent = JsonConverterMapper.getMappingForType(parentType);
        }
    };
}

function jsonProperty(serializedName, type) {
    return function (cls, property) {
        if (!type) {
            type = Reflect.getMetadata('design:type', cls, property);
        }
        var _a = normalizeConverterAndArgs(type), converter = _a.converter, args = _a.args;
        var typeMapping = JsonConverterMapper.createMappingForType(cls.constructor);
        var propertyMapping = {
            args: args,
            converter: converter,
            name: property,
            serializedName: serializedName,
        };
        typeMapping.properties.push(propertyMapping);
    };
}

export { EnumStrategy, JsonConverter, JsonConverterError, JsonConverterMapper, any, anyConverter, arrayConverter, arrayOf, autodetectConverterAndArgs, booleanConverter, enumConverter, enumOf, isBoolean, isNullOrUndefined, isNumber, isObject, isString, jsonObject, jsonProperty, keyValueConverter, keyValueOf, normalizeConverterAndArgs, numberConverter, objectConverter, stringConverter, tryParseBoolean, tryParseNumber };
